/*
 * Copyright 2022 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wSolver.h"
#include "wTimer.h"
#include "wProblem.h"
#include "wMedium.h"
#include "wuDirichlet.h"
#include "wuNeumann.h"
#include "wTDirichlet.h"
#include "wTNeumann.h"
#include "wTSource.h"
#include "wMSurface.h"
#include "wThermoMecaTerm.h"
#include "wTag.h"
#include "wMshData.h"
#include "wElement.h"
#include "wNode.h"
#include "wANSYSSolution.h"
#include <map>
#include <math.h>
#include "wResults.h"
#include "wMshExport.h"
#include "wLinearSolver.h"

#include <Eigen/Sparse>
#include <deque>

#include <tbb/global_control.h>
#include <tbb/parallel_for_each.h>
#include <tbb/spin_mutex.h>

#include <cstring>
#include <sstream>
#include <iostream>
/*
// [RB] not needed anymore
#ifdef max
#undef max    // pour std::numeric_limits<std::streamsize>::max()
#endif
*/
using namespace fwk;
using namespace tbox;
using namespace mirrors;

Solver::Solver(Problem &_pbl, std::shared_ptr<tbox::LinearSolver> _linsol, int _nthreads) : pbl(_pbl), linsol(_linsol), nthreads(_nthreads)
{
    // Update element memory
    pbl.initElems();
}

void Solver::write(std::ostream &out) const
{
    out << "mirrors::Solver:\n";
    out << pbl;
}

void Solver::start(MshExport *mshWriter)
{
    tms["total"].start();

    std::cout << "\nSolver::start()...\n";

    tbb::spin_mutex mutex;

    tbb::global_control control(tbb::global_control::max_allowed_parallelism, nthreads);

    //int a = 0;
    //int b = 10;
    //tbb::parallel_for_each(a,b, [=](int i) {if(i>5) return; std::cout << i << "\n";});

    //std::cout << "hi\n";
    //return;

    // ------------------------------------------
    //
    //               Preprocessing
    //
    // ------------------------------------------
    tms["prepro"].start();
    std::cout << "\tPreprocessing...\n";
    double T_ref = pbl.T_ref;

    auto msh = pbl.msh;

    size_t pbl_size;
    int T_first_index;

    if (pbl.problem_type == "mechanical")
    {
        pbl_size = 3 * (msh->nodes.size());
        T_first_index = 0;
    }
    else if (pbl.problem_type == "thermomechanical")
    {
        pbl_size = 4 * (msh->nodes.size());
        T_first_index = 3 * static_cast<int>(msh->nodes.size());
    }
    else if (pbl.problem_type == "thermal")
    {
        pbl_size = msh->nodes.size();
        T_first_index = 0;
    }
    else
        throw std::runtime_error("Unknown problem type.");

    tms["prepro"].stop();

    // ------------------------------------------
    //
    //            Matrices assembly
    //
    // ------------------------------------------
    tms["Kass"].start();
    std::cout << "\tAssembly...\n";

    // memory allocation
    std::deque<Eigen::Triplet<double>> Trip; // list of triplets to build stiffness matrix
    Eigen::SparseMatrix<double, Eigen::RowMajor> K(pbl_size, pbl_size);
    std::vector<double> rhs(pbl_size);
    std::vector<double> u(pbl_size);
    Eigen::Map<Eigen::VectorXd> u_(u.data(), u.size()), rhs_(rhs.data(), rhs.size());

    // loop on the media
    for (auto m : pbl.media)
    {
        Tag &current_tag = *(m->tag);
        std::cout << "\t\tProcessing " << current_tag << '\n';

        if (pbl.problem_type == "mechanical" || pbl.problem_type == "thermomechanical")
        {
            // H matrix
            double E = m->E;
            double nu = m->nu;

            Eigen::MatrixXd H = Eigen::MatrixXd::Zero(6, 6);
            H(0, 0) = 1 - nu;
            H(1, 1) = 1 - nu;
            H(2, 2) = 1 - nu;
            H(3, 3) = (1 - 2 * nu) / 2;
            H(4, 4) = (1 - 2 * nu) / 2;
            H(5, 5) = (1 - 2 * nu) / 2;
            H(0, 1) = nu;
            H(1, 0) = H(0, 1);
            H(0, 2) = nu;
            H(2, 0) = H(0, 2);
            H(1, 2) = nu;
            H(2, 1) = H(1, 2);
            H *= E / ((1 + nu) * (1 - 2 * nu));

            // K matrix
            //=============================tbb==================================
            tbb::parallel_for_each(current_tag.elems.begin(), current_tag.elems.end(), [&](Element *e) {
                //      Element *e = current_tag.elems[index];

                if (e->type() != ELTYPE::HEX8 && e->type() != ELTYPE::TETRA4)
                    return; //if the element is not a tetrahedron or a hexahedron, it is not treated

                Eigen::MatrixXd K_e = ThermoMecaTerm::buildK(*e, H);

                for (size_t ii = 0; ii < e->nodes.size(); ++ii)
                {
                    Node *nodi = e->nodes[ii];
                    for (size_t jj = 0; jj < e->nodes.size(); ++jj)
                    {
                        Node *nodj = e->nodes[jj];
                        //=====================================//
                        tbb::spin_mutex::scoped_lock lock(mutex);
                        //=====================================//
                        for (auto iii = 0; iii < 3; ++iii)
                            for (auto jjj = 0; jjj < 3; ++jjj)
                                if (K_e(3 * ii + iii, 3 * jj + jjj) != 0)
                                    Trip.push_back(Eigen::Triplet<double>(3 * (nodi->row) + iii, 3 * (nodj->row) + jjj, K_e(3 * ii + iii, 3 * jj + jjj)));
                    }
                }
            });
            //=============================tbb==================================
        }

        if (pbl.problem_type == "thermal" || pbl.problem_type == "thermomechanical")
        {
            // K_conductivity matrix
            double conductivity = m->k;
            Eigen::Matrix3d K_conductivity = Eigen::Matrix3d::Zero();
            for (auto i = 0; i < 3; ++i)
                K_conductivity(i, i) = conductivity;

            // L matrix
            //=============================tbb==================================
            tbb::parallel_for_each(current_tag.elems.begin(), current_tag.elems.end(), [&](Element *e) {
                if (e->type() != ELTYPE::HEX8 && e->type() != ELTYPE::TETRA4)
                    return; //if the element is not a tetrahedron or a hexahedron, it is not treated

                Eigen::MatrixXd L_e = ThermoMecaTerm::buildL(*e, K_conductivity);

                for (size_t ii = 0; ii < e->nodes.size(); ++ii)
                {
                    Node *nodi = e->nodes[ii];
                    for (size_t jj = 0; jj < e->nodes.size(); ++jj)
                    {
                        Node *nodj = e->nodes[jj];
                        //=====================================//
                        tbb::spin_mutex::scoped_lock lock(mutex);
                        //=====================================//
                        if (L_e(ii, jj) != 0)
                            Trip.push_back(Eigen::Triplet<double>(T_first_index + (nodi->row), T_first_index + (nodj->row), L_e(ii, jj)));
                    }
                }
            });
            //=============================tbb==================================
        }

        if (pbl.problem_type == "thermomechanical")
        {
            // D matrix
            double E = m->E;
            double nu = m->nu;
            double beta = m->beta;

            double lambda = (nu * E) / ((1 + nu) * (1 - 2 * nu));
            double G = E / (2 * (1 + nu));

            Eigen::Matrix3d D = Eigen::Matrix3d::Zero();
            //std::cout << "====================\n" << (lambda+(2./3.)*G) << "\n====================\n";
            for (auto i = 0; i < 3; ++i)
                D(i, i) = -3 * (lambda + (2. / 3.) * G) * beta;

            // S matrix
            //=============================tbb==================================
            tbb::parallel_for_each(current_tag.elems.begin(), current_tag.elems.end(), [&](Element *e) {
                if (e->type() != ELTYPE::HEX8 && e->type() != ELTYPE::TETRA4)
                    return; //if the element is not a tetrahedron or a hexahedron, it is not treated

                Eigen::MatrixXd S_e = ThermoMecaTerm::buildS(*e, D);

                for (size_t ii = 0; ii < e->nodes.size(); ++ii)
                {
                    Node *nodi = e->nodes[ii];
                    for (size_t jj = 0; jj < e->nodes.size(); ++jj)
                    {
                        Node *nodj = e->nodes[jj];
                        //=====================================//
                        tbb::spin_mutex::scoped_lock lock(mutex);
                        //=====================================//
                        for (int iii = 0; iii < 3; ++iii)
                            if (S_e(3 * ii + iii, jj) != 0)
                                Trip.push_back(Eigen::Triplet<double>(3 * (nodi->row) + iii, T_first_index + (nodj->row), S_e(3 * ii + iii, jj)));
                    }
                }
            });
            //=============================tbb==================================
        }
    }
    // Build matrix without BCs
    K.setFromTriplets(Trip.begin(), Trip.end());
    tms["Kass"].stop();

    // ------------------------------------------
    //
    //     volumetric heat sources and forces
    //
    // ------------------------------------------
    tms["vol_hs_F"].start();
    std::cout << "\tSources...\n";

    if (pbl.problem_type == "thermal" || pbl.problem_type == "thermomechanical")
    {
        // loop on the thermal volumetric heat
        for (auto b : pbl.Tsrcs)
        {
            Tag &current_tag = *(b->tag);
            b->integrated_values = 0;
            //loop on the elements
            for (auto e : current_tag.elems)
            {
                if (e->type() != ELTYPE::HEX8 && e->type() != ELTYPE::TETRA4)
                    continue; //if the element is not a tetrahedron or a hexahedron, it is not treated

                Eigen::MatrixXd N_e = ThermoMecaTerm::buildM(*e);

                for (size_t ii = 0; ii < e->nodes.size(); ++ii)
                {
                    //double tmp_value = 0;
                    Node *nodi = e->nodes[ii];
                    for (size_t jj = 0; jj < e->nodes.size(); ++jj)
                        if (N_e(ii, jj) != 0)
                        {
                            rhs[T_first_index + (nodi->row)] += (b->values) * N_e(ii, jj);
                            b->integrated_values += (b->values) * N_e(ii, jj);
                            //        tmp_value += N_e(ii,jj);
                        }
                    //std::cout << tmp_value << "\n";
                }

                //std::cout << "--------------\nElement " << *e << "\nCurrent integrated values: "<< b->integrated_values << "\n";
            }
            std::cout << b->THS_name << " integrated values:" << b->integrated_values << "\n";
        }
    }
    tms["vol_hs_F"].stop();

    // ------------------------------------------
    //
    //       von Neumann boundary conditions
    //
    // ------------------------------------------
    tms["nBC"].start();
    std::cout << "\tBoundary conditions...\n";
    if (pbl.problem_type == "mechanical" || pbl.problem_type == "thermomechanical")
    {
        // loop on the mechanical von Neumann BC
        for (auto b : pbl.nuBCs)
        {
            Tag &current_tag = *(b->tag);

            //loop on the elements
            for (auto e : current_tag.elems)
            {
                if (e->type() != ELTYPE::HEX8 && e->type() != ELTYPE::TETRA4)
                    continue; //if the element is not a triangle or a quadrangle, it is not treated

                Eigen::MatrixXd N_e = ThermoMecaTerm::buildM(*e);

                for (size_t ii = 0; ii < e->nodes.size(); ++ii)
                {
                    Node *nodi = e->nodes[ii];
                    for (size_t jj = 0; jj < e->nodes.size(); ++jj)
                        if (N_e(ii, jj) != 0)
                            for (auto iii = 0; iii < 3; ++iii)
                                if (b->which_dof(iii) == 1)
                                    rhs[3 * (nodi->row) + iii] += b->values(iii) * N_e(ii, jj);
                }
            }
        }
    }
    if (pbl.problem_type == "thermal" || pbl.problem_type == "thermomechanical")
    {
        // loop on the thermal von Neumann BC
        for (auto b : pbl.nTBCs)
        {
            Tag &current_tag = *(b->tag);
            b->integrated_values = 0;
            //loop on the elements
            for (auto e : current_tag.elems)
            {
                if (e->type() != ELTYPE::HEX8 && e->type() != ELTYPE::TETRA4)
                    continue; //if the element is not a triangle or a quadrangle, it is not treated

                Eigen::MatrixXd N_e = ThermoMecaTerm::buildM(*e);

                for (size_t ii = 0; ii < e->nodes.size(); ++ii)
                {
                    Node *nodi = e->nodes[ii];
                    for (size_t jj = 0; jj < e->nodes.size(); ++jj)
                        if (N_e(ii, jj) != 0)
                        {
                            rhs[T_first_index + (nodi->row)] += (b->values) * N_e(ii, jj);
                            b->integrated_values += (b->values) * N_e(ii, jj);
                        }
                }
            }
            std::cout << b->BC_name << " integrated values:" << b->integrated_values << "\n";
        }
    }
    tms["nBC"].stop();

    // ------------------------------------------
    //
    //        Dirichlet boundary conditions
    //
    // ------------------------------------------
    tms["dBC"].start();
    if (pbl.problem_type == "mechanical" || pbl.problem_type == "thermomechanical")
    {
        // loop on the mechanical Dirichlet BC
        for (auto b : pbl.duBCs)
        {
            Tag &current_tag = *(b->tag);

            //loop on the elements
            for (auto e : current_tag.elems)
            {
                if (e->type() == ELTYPE::QUAD4)
                    ;
                else if (e->type() == ELTYPE::TRI3)
                    ;
                else if (e->type() == ELTYPE::LINE2)
                    ;
                else
                    continue; //if the element is not a triangle or a quadrangle, it is not treated

                for (size_t ii = 0; ii < e->nodes.size(); ++ii)
                    for (auto iii = 0; iii < 3; ++iii)
                        if (b->which_dof(iii) == 1)
                        {
                            Node *nodi = e->nodes[ii];
                            for (Eigen::SparseMatrix<double, Eigen::RowMajor>::InnerIterator it(K, 3 * nodi->row + iii); it; ++it)
                            {
                                if (it.row() == it.col())
                                    it.valueRef() = 1.;
                                else
                                    it.valueRef() = 0.;
                            }
                            rhs[3 * (nodi->row) + iii] = b->values(iii);
                        }
            }
        }
    }
    if (pbl.problem_type == "thermal" || pbl.problem_type == "thermomechanical")
    {
        // loop on the thermal Dirichlet BC
        for (auto b : pbl.dTBCs)
        {
            Tag &current_tag = *(b->tag);

            //loop on the elements
            for (auto e : current_tag.elems)
            {
                if (e->type() == ELTYPE::QUAD4)
                    ;
                else if (e->type() == ELTYPE::TRI3)
                    ;
                else
                    continue; //if the element is not a triangle or a quadrangle, it is not treated

                for (size_t ii = 0; ii < e->nodes.size(); ++ii)
                {
                    Node *nodi = e->nodes[ii];
                    for (Eigen::SparseMatrix<double, Eigen::RowMajor>::InnerIterator it(K, T_first_index + nodi->row); it; ++it)
                    {
                        if (it.row() == it.col())
                            it.valueRef() = 1.;
                        else
                            it.valueRef() = 0.;
                    }
                    rhs[T_first_index + nodi->row] = b->values - T_ref;
                }
            }
        }
    }
    tms["dBC"].stop();

    // ------------------------------------------
    //
    //              Matrices cleaning
    //
    // ------------------------------------------
    tms["Kclean"].start();
    std::cout << "\tCleaning matrices...\n";
    // Clean matrix and turn to compressed row format
    K.prune(0.);
    K.makeCompressed();
    tms["Kclean"].stop();

    // ------------------------------------------
    //
    //                  Solver
    //
    // ------------------------------------------
    tms["solve"].start();
    std::cout << "\tSolving linear system...\n";
    linsol->compute(K, rhs_, u_);
    tms["solve"].stop();

    // ------------------------------------------
    //
    //               Postprocessing
    //
    // ------------------------------------------
    tms["post"].start();
    std::cout << "\tPostprocessing...\n";
    // setup results
    Results results;

    std::vector<double> displacement_x(msh->nodes.size());
    std::vector<double> displacement_y(msh->nodes.size());
    std::vector<double> displacement_z(msh->nodes.size());
    std::vector<double> displacement_u(msh->nodes.size());

    std::vector<std::vector<double>> ansys_tmp(pbl.ANSYSSol.size(), std::vector<double>(msh->nodes.size(), 0));
    std::vector<std::vector<double>> ansys_tmp2(pbl.ANSYSSol.size(), std::vector<double>(msh->nodes.size(), 0));
    std::vector<std::vector<double>> ansys_tmp3(pbl.ANSYSSol.size(), std::vector<double>(msh->nodes.size(), 0));

    std::vector<double> n_displacement(msh->nodes.size());
    std::vector<double> r_displacement(msh->nodes.size());
    std::vector<double> distance(msh->nodes.size());
    std::vector<double> rho(msh->nodes.size());
    std::vector<double> theta(msh->nodes.size());
    std::vector<double> surface(msh->nodes.size());
    std::vector<double> x_mirror(msh->nodes.size());
    std::vector<double> y_mirror(msh->nodes.size());

    std::vector<double> Z_1(msh->nodes.size());
    std::vector<double> Z_2(msh->nodes.size());
    std::vector<double> Z_3(msh->nodes.size());
    std::vector<double> Z_4(msh->nodes.size());
    std::vector<double> Z_5(msh->nodes.size());
    std::vector<double> Z_6(msh->nodes.size());
    std::vector<double> Z_7(msh->nodes.size());
    std::vector<double> Z_8(msh->nodes.size());
    std::vector<double> Z_9(msh->nodes.size());
    std::vector<double> Z_10(msh->nodes.size());

    std::vector<Eigen::Vector3d> displacement(msh->nodes.size());
    std::vector<Eigen::Vector3d> norm(msh->nodes.size());

    std::vector<Eigen::MatrixXd> Strain_vector(msh->elems.size());
    for (auto &k : Strain_vector)
        k = Eigen::Matrix3d::Zero();
    std::vector<Eigen::MatrixXd> Stress_vector(msh->elems.size());
    for (auto &k : Stress_vector)
        k = Eigen::Matrix3d::Zero();

    std::vector<double> Strain_xx_vector(msh->elems.size());
    std::vector<double> Strain_xy_vector(msh->elems.size());
    std::vector<double> Strain_xz_vector(msh->elems.size());
    std::vector<double> Strain_yy_vector(msh->elems.size());
    std::vector<double> Strain_yz_vector(msh->elems.size());
    std::vector<double> Strain_zz_vector(msh->elems.size());

    std::vector<double> Stress_xx_vector(msh->elems.size());
    std::vector<double> Stress_xy_vector(msh->elems.size());
    std::vector<double> Stress_xz_vector(msh->elems.size());
    std::vector<double> Stress_yy_vector(msh->elems.size());
    std::vector<double> Stress_yz_vector(msh->elems.size());
    std::vector<double> Stress_zz_vector(msh->elems.size());

    std::vector<double> T(msh->nodes.size());

    double T_max = 0., u_max = 0., x_max = 0., y_max = 0., z_max = 0.;

    if (pbl.problem_type == "mechanical" || pbl.problem_type == "thermomechanical")
    {
        results.scalars_at_nodes["x"] = &displacement_x;
        results.scalars_at_nodes["y"] = &displacement_y;
        results.scalars_at_nodes["z"] = &displacement_z;
        results.scalars_at_nodes["u"] = &displacement_u;

        results.scalars_at_elems["Strain[xx]"] = &Strain_xx_vector;
        results.scalars_at_elems["Strain[xy]"] = &Strain_xy_vector;
        results.scalars_at_elems["Strain[xz]"] = &Strain_xz_vector;
        results.scalars_at_elems["Strain[yy]"] = &Strain_yy_vector;
        results.scalars_at_elems["Strain[yz]"] = &Strain_yz_vector;
        results.scalars_at_elems["Strain[zz]"] = &Strain_zz_vector;

        results.scalars_at_elems["Stress[xx]"] = &Stress_xx_vector;
        results.scalars_at_elems["Stress[xy]"] = &Stress_xy_vector;
        results.scalars_at_elems["Stress[xz]"] = &Stress_xz_vector;
        results.scalars_at_elems["Stress[yy]"] = &Stress_yy_vector;
        results.scalars_at_elems["Stress[yz]"] = &Stress_yz_vector;
        results.scalars_at_elems["Stress[zz]"] = &Stress_zz_vector;

        results.vectors_at_nodes["displacement"] = &displacement;

        results.tensors_at_elems["Strain[tensor]"] = &Strain_vector;
        results.tensors_at_elems["Stress[tensor]"] = &Stress_vector;

        for (size_t i = 0; i < msh->nodes.size(); ++i)
        {
            displacement_x[i] = u[i * 3];
            if (x_max < fabs(displacement_x[i]))
                x_max = fabs(displacement_x[i]);
            displacement_y[i] = u[i * 3 + 1];
            if (y_max < fabs(displacement_y[i]))
                y_max = fabs(displacement_y[i]);
            displacement_z[i] = u[i * 3 + 2];
            if (z_max < fabs(displacement_z[i]))
                z_max = fabs(displacement_z[i]);
            displacement_u[i] = sqrt(u[i * 3] * u[i * 3] + u[i * 3 + 1] * u[i * 3 + 1] + u[i * 3 + 2] * u[i * 3 + 2]);
            if (u_max < fabs(displacement_u[i]))
                u_max = fabs(displacement_u[i]);

            displacement[i] = Eigen::Vector3d(displacement_x[i], displacement_y[i], displacement_z[i]);
        }

        // normal displacement of the mirror surface
        // + compute an equivalent surface for each of its node
        tms["Zernike"].start();
        for (auto ms : pbl.MS)
        {
            results.scalars_at_nodes["n_displacement"] = &n_displacement;
            results.scalars_at_nodes["r_displacement"] = &r_displacement;
            results.scalars_at_nodes["surface"] = &surface;
            results.scalars_at_nodes["distance"] = &distance;
            results.scalars_at_nodes["rho"] = &rho;
            results.scalars_at_nodes["theta"] = &theta;
            results.scalars_at_nodes["x_mirror"] = &x_mirror;
            results.scalars_at_nodes["y_mirror"] = &y_mirror;
            results.vectors_at_nodes["norm"] = &norm;

            results.scalars_at_nodes["Z_1"] = &Z_1;
            results.scalars_at_nodes["Z_2"] = &Z_2;
            results.scalars_at_nodes["Z_3"] = &Z_3;
            results.scalars_at_nodes["Z_4"] = &Z_4;
            results.scalars_at_nodes["Z_5"] = &Z_5;
            results.scalars_at_nodes["Z_6"] = &Z_6;
            results.scalars_at_nodes["Z_7"] = &Z_7;
            results.scalars_at_nodes["Z_8"] = &Z_8;
            results.scalars_at_nodes["Z_9"] = &Z_9;
            results.scalars_at_nodes["Z_10"] = &Z_10;

            Tag &current_tag = *(ms->tag);

            Eigen::Vector3d normal = current_tag.elems[0]->getNormal();

            double max_distance = 0.;

            for (auto e : current_tag.elems)
            {
                if (e->type() != ELTYPE::HEX8 && e->type() != ELTYPE::TETRA4)
                    continue; //if the element is not a triangle or a quadrangle, it is not treated

                Eigen::MatrixXd N_e = ThermoMecaTerm::buildM(*e);

                for (size_t ii = 0; ii < e->nodes.size(); ++ii)
                {
                    Node *nodi = e->nodes[ii];
                    n_displacement[nodi->row] = displacement[nodi->row].dot(normal);
                    r_displacement[nodi->row] = n_displacement[nodi->row];
                    norm[nodi->row] = normal;
                    distance[nodi->row] = (nodi->pos - ms->x0).norm();

                    if (max_distance < distance[nodi->row])
                        max_distance = distance[nodi->row];

                    rho[nodi->row] = (nodi->pos - ms->x0).norm();
                    x_mirror[nodi->row] = (nodi->pos - ms->x0).dot(ms->x1) / (ms->x1.norm());
                    y_mirror[nodi->row] = (nodi->pos - ms->x0).dot(ms->x2) / (ms->x2.norm());

                    Z_1[nodi->row] = 1.;

                    if (x_mirror[nodi->row] >= 0. && y_mirror[nodi->row] >= 0.)
                        theta[nodi->row] = atan(y_mirror[nodi->row] / x_mirror[nodi->row]);
                    else if (y_mirror[nodi->row] >= 0.)
                        theta[nodi->row] = atan(y_mirror[nodi->row] / x_mirror[nodi->row]) + M_PI;
                    else if (x_mirror[nodi->row] < 0.)
                        theta[nodi->row] = atan(y_mirror[nodi->row] / x_mirror[nodi->row]) + M_PI;
                    else
                        theta[nodi->row] = atan(y_mirror[nodi->row] / x_mirror[nodi->row]) + 2 * M_PI;

                    for (size_t jj = 0; jj < e->nodes.size(); ++jj)
                        surface[nodi->row] += N_e(ii, jj);
                }
            }

            for (size_t ii = 0; ii < msh->nodes.size(); ++ii)
            {
                rho[ii] /= max_distance;

                Z_2[ii] = 2 * rho[ii] * cos(theta[ii]);
                Z_3[ii] = 2 * rho[ii] * sin(theta[ii]);
                Z_4[ii] = sqrt(3) * (2 * pow(rho[ii], 2.) - 1);
                Z_5[ii] = sqrt(6) * pow(rho[ii], 2.) * sin(2 * theta[ii]);
                Z_6[ii] = sqrt(6) * pow(rho[ii], 2.) * cos(2 * theta[ii]);
                Z_7[ii] = sqrt(8) * (3 * pow(rho[ii], 3.) - 2 * rho[ii]) * sin(theta[ii]);
                Z_8[ii] = sqrt(8) * (3 * pow(rho[ii], 3.) - 2 * rho[ii]) * cos(theta[ii]);
                Z_9[ii] = sqrt(8) * pow(rho[ii], 3.) * sin(3 * theta[ii]);
                Z_10[ii] = sqrt(8) * pow(rho[ii], 3.) * cos(3 * theta[ii]);
            }

            Eigen::MatrixXd W = Eigen::MatrixXd::Zero(msh->nodes.size(), msh->nodes.size());
            Eigen::MatrixXd Z(msh->nodes.size(), 10);
            Eigen::MatrixXd A(10, 10);
            Eigen::VectorXd rhs_Z(10); // RHS

            for (size_t ii = 0; ii < msh->nodes.size(); ++ii)
            {
                W(ii, ii) = surface[ii];

                Z(ii, 0) = Z_1[ii];
                Z(ii, 1) = Z_2[ii];
                Z(ii, 2) = Z_3[ii];
                Z(ii, 3) = Z_4[ii];
                Z(ii, 4) = Z_5[ii];
                Z(ii, 5) = Z_6[ii];
                Z(ii, 6) = Z_7[ii];
                Z(ii, 7) = Z_8[ii];
                Z(ii, 8) = Z_9[ii];
                Z(ii, 9) = Z_10[ii];
            }

            A = Z.transpose() * W * Z;
            rhs_Z = Z.transpose() * W * Eigen::Map<Eigen::VectorXd>(n_displacement.data(), msh->nodes.size());
            Eigen::VectorXd c_Z = A.partialPivLu().solve(rhs_Z);

            std::cout << c_Z;

            for (size_t ii = 0; ii < msh->nodes.size(); ++ii)
                for (auto jj = 0; jj < 10; ++jj)
                    r_displacement[ii] -= c_Z[jj] * Z(ii, jj);
        }
        tms["Zernike"].stop();

        Eigen::MatrixXd Strain;
        Eigen::MatrixXd Sigma;

        // loop on the medium
        double sigma_max = 0.;
        double sigma_vm = 0.;
        for (auto m : pbl.media)
        {
            Tag &current_tag = *(m->tag);

            double E = m->E;
            double nu = m->nu;

            double lambda = (nu * E) / ((1 + nu) * (1 - 2 * nu));
            double G = E / (2 * (1 + nu));

            Eigen::MatrixXd H = Eigen::MatrixXd::Zero(9, 9);

            for (auto i = 0; i < 3; ++i)
                for (auto j = 0; j < 3; ++j)
                    for (auto k = 0; k < 3; ++k)
                        for (auto l = 0; l < 3; ++l)
                        {
                            if (i == j && k == l)
                                H(i * 3 + j, k * 3 + l) = lambda;
                            if (i == k && j == l)
                                H(i * 3 + j, k * 3 + l) = H(i * 3 + j, k * 3 + l) + G;
                            if (i == l && j == k)
                                H(i * 3 + j, k * 3 + l) = H(i * 3 + j, k * 3 + l) + G;
                        }

            for (auto e : current_tag.elems)
            {
                if (e->type() == ELTYPE::HEX8 || e->type() == ELTYPE::TETRA4)
                    ThermoMecaTerm::strain_stress(*e, H, u, Strain, Sigma);
                else
                    continue;

                auto i = e->no - 1;

                Strain_vector[i] = Strain;
                Stress_vector[i] = Sigma;

                Stress_xx_vector[i] = Sigma(0, 0);
                Stress_xy_vector[i] = Sigma(0, 1);
                Stress_xz_vector[i] = Sigma(0, 2);
                Stress_yy_vector[i] = Sigma(1, 1);
                Stress_yz_vector[i] = Sigma(1, 2);
                Stress_zz_vector[i] = Sigma(2, 2);

                Strain_xx_vector[i] = Strain(0, 0);
                Strain_xy_vector[i] = Strain(0, 1);
                Strain_xz_vector[i] = Strain(0, 2);
                Strain_yy_vector[i] = Strain(1, 1);
                Strain_yz_vector[i] = Strain(1, 2);
                Strain_zz_vector[i] = Strain(2, 2);

                sigma_vm = sqrt((pow(Sigma(0, 0) - Sigma(1, 1), 2.) + pow(Sigma(1, 1) - Sigma(2, 2), 2.) + pow(Sigma(2, 2) - Sigma(0, 0), 2.) + 6 * (pow(Sigma(0, 1), 2.) + pow(Sigma(1, 2), 2.) + pow(Sigma(2, 0), 2.))) / 2);

                if (sigma_max < sigma_vm)
                    sigma_max = sigma_vm;
            }
        }
        pbl.Sol = sigma_max;
    }
    if (pbl.problem_type == "thermal" || pbl.problem_type == "thermomechanical")
    {
        results.scalars_at_nodes["T"] = &T;

        for (size_t i = 0; i < msh->nodes.size(); ++i)
        {
            T[i] = u[T_first_index + i] + T_ref;
            if (T_max < T[i])
                T_max = T[i];
        }
    }

    int as_index = 0;
    for (auto AS : pbl.ANSYSSol)
    {
        if (AS->solution_type == "T" && pbl.problem_type == "mechanical")
            continue;
        if (AS->solution_type == "u" && pbl.problem_type == "thermal")
            continue;
        if (AS->solution_type == "x" && pbl.problem_type == "thermal")
            continue;
        if (AS->solution_type == "y" && pbl.problem_type == "thermal")
            continue;
        if (AS->solution_type == "z" && pbl.problem_type == "thermal")
            continue;

        results.scalars_at_nodes[AS->solution_name] = &ansys_tmp[as_index];
        std::string name2 = AS->solution_name + "_diff_abs";
        results.scalars_at_nodes[name2] = &ansys_tmp2[as_index];
        name2 = AS->solution_name + "_diff_rel";
        results.scalars_at_nodes[name2] = &ansys_tmp3[as_index];

        FILE *read_file = fopen(AS->solution_file.c_str(), "r");
        if (!read_file)
        {
            std::stringstream str;
            str << "Error opening file \"" << AS->solution_file << "\"";
            throw std::runtime_error(str.str());
        }
        double max_diff_abs = 0;
        double max_diff_rel = 0;

        int index;

        while (1)
        {
            // read next line
            char line[128];
            if (fgets(line, sizeof line, read_file) == NULL)
                break;
            else if (line[0] != 'N')
            {
                std::istringstream input(line);

                int i = 0;

                int i1 = -1;
                double n1;
                for (;;)
                {
                    double n;
                    input >> n;

                    if (input.eof() || input.bad())
                        break;
                    else if (input.fail())
                    {
                        input.clear();                                                   // unset failbit
                        input.ignore(std::numeric_limits<std::streamsize>::max(), '\n'); // skip bad input
                    }
                    else if (i == 0)
                        i1 = static_cast<int>(n); // c'est pas beau! (Romain)
                    else if (i == 1)
                        n1 = n;

                    ++i;
                }
                if (i1 != -1)
                {
                    index = i1 - 1;
                    ansys_tmp[as_index][index] = n1;
                    if (AS->solution_type == "T")
                    {
                        ansys_tmp2[as_index][index] = fabs(ansys_tmp[as_index][index] - T[index]);
                        ansys_tmp3[as_index][index] = ansys_tmp2[as_index][index] / T_max;
                    }
                    else if (AS->solution_type == "u")
                    {
                        ansys_tmp2[as_index][index] = fabs(ansys_tmp[as_index][index] - displacement_u[index]);
                        ansys_tmp3[as_index][index] = ansys_tmp2[as_index][index] / u_max;
                    }
                    else if (AS->solution_type == "x")
                    {
                        ansys_tmp2[as_index][index] = fabs(ansys_tmp[as_index][index] - displacement_x[index]);
                        ansys_tmp3[as_index][index] = ansys_tmp2[as_index][index] / x_max;
                    }
                    else if (AS->solution_type == "y")
                    {
                        ansys_tmp2[as_index][index] = fabs(ansys_tmp[as_index][index] - displacement_y[index]);
                        ansys_tmp3[as_index][index] = ansys_tmp2[as_index][index] / y_max;
                    }
                    else if (AS->solution_type == "z")
                    {
                        ansys_tmp2[as_index][index] = fabs(ansys_tmp[as_index][index] - displacement_z[index]);
                        ansys_tmp3[as_index][index] = ansys_tmp2[as_index][index] / z_max;
                    }

                    if (max_diff_abs <= ansys_tmp2[as_index][index])
                        max_diff_abs = ansys_tmp2[as_index][index];

                    if (max_diff_rel <= ansys_tmp3[as_index][index])
                        max_diff_rel = ansys_tmp3[as_index][index];
                }
            }
        }

        fclose(read_file);
        AS->diff_abs = max_diff_abs;
        AS->diff_rel = max_diff_rel;
        ++as_index;
    }

    mshWriter->save(msh->name, results);
    tms["post"].stop();
    std::cout << "\tdone.\n";

    tms["total"].stop();

    std::cout << "\n---CPU statistics ---\n";
    std::cout << tms;
    std::cout << "---------------------\n";
}
