/*
 * Copyright 2022 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain k copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wThermoMecaTerm.h"

#include "wElement.h"
#include "wCache.h"
#include "wGauss.h"
#include "wNode.h"

using namespace mirrors;
using namespace tbox;

/**
 * @brief Build volume/surface mass matrix for the thermo-mechanical equation on one element
 */
Eigen::MatrixXd ThermoMecaTerm::buildM(Element const &e)
{
    // Get precomputed values
    Cache &cache = e.getVCache();
    Gauss &gauss = cache.getVGauss();

    // Mass matrix
    Eigen::MatrixXd M = Eigen::MatrixXd::Zero(e.nodes.size(), e.nodes.size());
    for (size_t k = 0; k < gauss.getN(); ++k)
    {
        // Shape functions
        Eigen::VectorXd const &ff = cache.getSf(k);
        M += ff * ff.transpose() * gauss.getW(k) * e.getDetJ(k);
    }
    return M;
}

/**
 * @brief Build mechanical matrix for the thermo-mechanical equation on one element
 */
Eigen::MatrixXd ThermoMecaTerm::buildK(tbox::Element const &e, Eigen::MatrixXd const &H)
{
    // Get precomputed values
    Cache &cache = e.getVCache();
    Gauss &gauss = cache.getVGauss();
    size_t ns = e.nodes.size(); // number of nodes

    // Meca matrix
    Eigen::MatrixXd K = Eigen::MatrixXd::Zero(3 * ns, 3 * ns); // 3 dimensions
    for (size_t k = 0; k < gauss.getN(); ++k)
    {
        // Jacobian inverse
        Eigen::Matrix3d const &invJ = e.getJinv(k);
        // Fill B matrix (shape functions)
        Eigen::MatrixXd const &dff = cache.getDsf(k);
        Eigen::MatrixXd B = Eigen::MatrixXd::Zero(6, 3 * ns);
        for (size_t i = 0; i < ns; ++i)
        {
            // Compute [Ni,x Ni,y, Ni_z]
            Eigen::Vector3d dN = invJ * dff.col(i);
            B(0, 3 * i) = dN(0);
            B(1, 3 * i + 1) = dN(1);
            B(2, 3 * i + 2) = dN(2);
            B(3, 3 * i + 1) = dN(2);
            B(3, 3 * i + 2) = dN(1);
            B(4, 3 * i) = dN(2);
            B(4, 3 * i + 2) = dN(0);
            B(5, 3 * i) = dN(1);
            B(5, 3 * i + 1) = dN(0);
        }

        // Compute stiffness matrix
        K += B.transpose() * H * B * gauss.getW(k) * e.getDetJ(k);
    }
    return K;
}

/**
 * @brief Build thermal matrix for the thermo-mechanical equation on one element
 */
Eigen::MatrixXd ThermoMecaTerm::buildL(Element const &e, Eigen::MatrixXd const &C)
{
    // Get precomputed values
    Cache &cache = e.getVCache();
    Gauss &gauss = cache.getVGauss();

    // Thermal matrix
    Eigen::MatrixXd L = Eigen::MatrixXd::Zero(e.nodes.size(), e.nodes.size());
    for (size_t k = 0; k < gauss.getN(); ++k)
    {
        // Jacobian
        double detJ = e.getDetJ(k);
        Eigen::Matrix3d const &J = e.getJinv(k);
        Eigen::MatrixXd const &dff = cache.getDsf(k);

        // [AC] I refactored this from an entry by entry assembly to a full matrix multiplication,
        // but as nothing tests this method, I could not check the results.
        L += (C * J * dff).transpose() * J * dff * gauss.getW(k) * detJ;
    }
    return L;
}

/**
 * @brief Build thermo-mechanical matrix for the thermo-mechanical equation on one element
 */
Eigen::MatrixXd ThermoMecaTerm::buildS(Element const &e, Eigen::MatrixXd const &D)
{
    // Get precomputed values
    Cache &cache = e.getVCache();
    Gauss &gauss = cache.getVGauss();
    size_t ns = e.nodes.size(); // number of nodes

    // TM matrix
    Eigen::MatrixXd S = Eigen::MatrixXd::Zero(3 * ns, ns); // 3 dimensions
    for (size_t k = 0; k < gauss.getN(); ++k)
    {
        // Jacobian and shape functions
        double detJ = e.getDetJ(k);
        Eigen::Matrix3d const &J = e.getJinv(k);
        Eigen::MatrixXd const &dff = cache.getDsf(k);

        // S
        for (size_t i = 0; i < ns; ++i)
            S.block(i * 3, 0, 3, ns) += D * J * dff.col(i) * cache.getSf(k).transpose() * gauss.getW(k) * detJ;
    }
    return S;
}

/**
 * @brief Compute strain and stresses on one element
 */
void ThermoMecaTerm::strain_stress(Element const &e, Eigen::MatrixXd const &H, std::vector<double> const &u, Eigen::MatrixXd &Strain, Eigen::MatrixXd &Stress)
{
    // Get precomputed values
    Cache &cache = e.getVCache();
    Gauss &gauss = cache.getVGauss();
    size_t ns = e.nodes.size(); // number of nodes

    Strain = Stress = Eigen::Matrix3d::Zero(); // 3 dimensions
    Eigen::MatrixXd u_e(3, ns);
    // store the solution
    for (size_t i = 0; i < ns; ++i)
    {
        u_e(0, i) = u[3 * (e.nodes[i]->row)];
        u_e(1, i) = u[3 * (e.nodes[i]->row) + 1];
        u_e(2, i) = u[3 * (e.nodes[i]->row) + 2];
    }
    for (size_t k = 0; k < gauss.getN(); ++k)
    {
        // Jacobian and strain
        Eigen::Matrix3d const &J = e.getJinv(k);
        Eigen::MatrixXd const &dff = cache.getDsf(k);
        Eigen::Matrix3d du_d_Phi = Eigen::Matrix3d::Zero();
        for (size_t i = 0; i < ns; ++i)
            du_d_Phi += dff.col(i) * u_e.col(i).transpose();
        Strain += ((J * du_d_Phi) + (J * du_d_Phi).transpose()) / ns / 2;
    }
    // stress
    for (size_t i = 0; i < 3; ++i)
        for (size_t j = 0; j < 3; ++j)
            for (size_t k = 0; k < 3; ++k)
                for (size_t l = 0; l < 3; ++l)
                    Stress(i, j) += H(i * 3 + j, k * 3 + l) * Strain(k, l);
}
